; example of initializing AP CPUs different way than stated in CPU manual
; we rely here on the fact, that firmware already initialized AP CPUs and later put them into HLT state with interrupts disabled from which they could be waken up by NMI
; I saw it somewhere long time ago when reversing a binary, but can't find it anymore so I tried quickly reconstruct the idea from my memories

; this code is executed in real mode at bootstrap CPU
; the BSP CPU hooks int2 (NMI) to point it to this address, switches into long mode and sends NMI to all application CPUs
; then all AP CPUs start to execute from the address in NMI vector 2 of IDT (this assumes that the AP CPUs were put into HLT state in realmode so they use realmode IDT)
; all CPUs print their APIC ID numbers and then enter HLT state (you must use reset or power button on your PC or unplug electric cable etd)

org 7C00h

use16

start16:
	xor	ax,ax
	mov	ds,ax
	mov	es,ax
	cli

; we will run the whole time without using stack and with interrupts are disabled, so there is no need to setup stack
;	mov	ss,ax
;	mov	sp,ax

	lgdt	[gdtr]



; hook int 2 (NMI) in realmode IDT
	mov	eax,start16
	xchg	[2*4],eax
	mov	[orig_rm_nmi_vector],eax



	mov	eax,cr4
	or	al,(1 shl 5)			; PAE bit 5.
	mov	cr4,eax 			; enable global pages and physical-address extensions

	mov	ecx,0C0000080h			; EFER
	rdmsr
	or	ah,1 shl (8-8)			; EFER.LME enable long mode
	wrmsr

	mov	eax,1000h
	mov	cr3,eax				; pointer to paging tables
 

	mov	ecx,1Bh				; IA32_APIC_BASE
	rdmsr
	test	ah,1 shl (8-8)			; BSP flag (RO) set to 1 for bootstrap CPU
	jz	after_preparing_paging_tables

	mov	di,1000h
; clear 3 pages
	mov	cx,6 * 1000h / 2		; clear all 6 pages used for constructing paging tables
	xor	ax,ax
	cld
	repz	stosw				; clear the page tables at memory 1000h-6FFFh
	mov	word [1000h],2000h + 111b	; PML4E[0] = pointer to first PDP table
	mov	word [2000h],3000h + 111b	; PDPE[0] = pointer to first page directory table
; now we map 2 MB of memory, the memory will be identity-mapped (physical address in RAM = virtual address), when paging will be enabled, physical address switches into the same virtual address thanks to identity-mapping
	mov	word [3000h],0 + 10000011b	; PDE[0]

after_preparing_paging_tables:
	mov	eax,cr0
	or	eax,(1 shl 31) or (1 shl 0)
	mov	cr0,eax 			; enable paging (bit 31.) and protected mode (bit 0.)
						; instruction that enables paging and the following instructions must be located in an identity-mapped page until such time that a branch to non-identity mapped pages can be effected

	jmp	code64_selector : start64

use64
start64:
; this protected mode code is executed by all CPUs at the same address, all application CPUs jump here by the last jmp code64_selector : start64

	mov	eax,data64_selector
;	mov	edx,STACK64_SELECTOR_init
	mov	ds,ax
	mov	es,ax

	mov	ecx,1Bh				; IA32_APIC_BASE
	rdmsr

	or	ah,1 shl (11-8)			; bit 11. APIC_ENABLE
	wrmsr

	mov	ebx,not 0FFFh
	and	ebx,eax
	shl	rdx,32

	lea	rbp,[rdx+rbx*1+11011b]

; rbp = PTE (PT entry) to be written into PT to map APIC to virtual memory = APIC base with low bits sets to match the PT entry (PCD=1 PWT=1 W=1 P=1)

apic_virtual_memory				= - 20h*1000h
a = (apic_virtual_memory shr 39) and 111111111b
b = (apic_virtual_memory shr 30) and 111111111b
c = (apic_virtual_memory shr 21) and 111111111b
d = (apic_virtual_memory shr 12) and 111111111b

	mov	rbx,apic_virtual_memory
; rbx = virtual memory where APIC is mapped

	test	ah,1 shl (8-8)			; BSP flag (RO) set to 1 for bootstrap CPU
	jz	application_CPU

bootstrap_CPU:
	mov	rax,cr3
; rax = 1000h now, N.B. the low 2 MB of memory is identity mapped (virtual memory = physical memory)
	and	ax,not (1000h - 1)		; wipe out low bits which may be set in CR3, currently we didn't set any of these bits in CR3 so no need to erase them
						; N.B. never use the instruction with EAX register as that operation wipes out bits 63-32 of RAX !!!

	mov	word [1000h + (rax-1000h) + a*8],4000h + 111b	; field 0. of of PML4 = pointer to first PDP table
					; upper dword was set 0 at the above repz stosw
	mov	word [4000h + (rax-1000h) + b*8],5000h + 111b	; field 0. of PDP table = pointer to first page directory table
	mov	word [5000h + (rax-1000h) + c*8],6000h + 111b	; field 0. of PDP table = pointer to first page directory table
	mov	[6000h + (rax-1000h) + d*8],rbp		; N.B. the above entries have bit 2. (U/S) set to 1, this last has set it to 0, so only supervisor (ring0, 1, 2) is allowed to access memory page mapped by last entry in PD (the entry has name PDE)




; bootstrap CPU needs to fire out some IPIs so activate AP CPUs, in this case NMI is used (AP were already activated by firmware which then put them into HLT state with interrupts disabled from which they could be woken up by INIT-SIPI but also by a NMI)
	mov	dword [rbx + 0300h],(11b shl 18) + (0 shl 15) + (1 shl 14) + (100b shl 8)
; Destination Shorthand = All excluding Self
; Trigger = Edge
; Level = Assert
; Delivery Mode = NMI
; Vector Bits 7-0 ignored when the message type is NMI


 
; jmp print_APIC_ID

application_CPU:
print_APIC_ID:
	mov	eax,[rbx + 0020h]
	shr	eax,24
	lea	edx,[rax + '0']			; transform 00h binary to ascii '0'
	or	edx,1F001F00h
application_CPU_loop:
	xor	edx,0FFFFFF00h
	mov	[0B8000h+rax*4],edx

;	mov	ecx,2000000000			; wait 500ms
;	loop	$
;	jmp	application_CPU_loop
finito:
cli
hlt
jmp finito



align 10h
gdt:		dq	0000000000000000h
sys_code64:	dq	00209A0000000000h
sys_data64:	dq	00CF96000000FFFFh
gdt_end:
code64_selector	=	sys_code64 - gdt
data64_selector	=	sys_data64 - gdt

gdtr:
		dw	gdt_end - gdt - 1
		dd	gdt


orig_rm_nmi_vector	dd	0

times 512-2-$+$$	db	0

org 7DFEh
	dw	0AA55h
